#!/usr/bin/python
# Written by Stephane Graber <stgraber@ubuntu.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# Common settings
verbose=3       # Verbosity level, 0 = nothing, 1 = critical, 2 = info, 3 = debug
tag=1           # Set to 0 to disable tagging (will print a message in info mode to the console instead)
lock=0          # Enable locking

# Database settings
db_host="/var/run/postgresql/"
db_user="db"
db_password=None
db_database="qatracker"

import sys, pgdb, re, os

# Check if lock is present
if lock:
    if os.path.exists("/tmp/lp-integration.lock"):
        sys.exit("lp-integration.py is already running")
    else:
        lock=open("/tmp/lp-integration.lock","w+")
        lock.write(str(os.getpid()))
        lock.close()

# Load launchpadlib
from launchpadlib.launchpad import Launchpad, EDGE_SERVICE_ROOT
from launchpadlib.credentials import Credentials
from launchpadlib.errors import HTTPError

# Initialize the cache
cachedir = ".launchpadlib/cache"
if not os.path.exists(cachedir):
    os.makedirs(cachedir,0700)

# Get credential
credfile = "lp-integration.cred"
try:
    credentials = Credentials()
    credentials.load(open(credfile))
    launchpad = Launchpad(credentials, EDGE_SERVICE_ROOT, cachedir)
except:
    launchpad = Launchpad.get_token_and_login("lp-integration", EDGE_SERVICE_ROOT, cachedir)
    launchpad.credentials.save(open(credfile,"w",0600)) 

# Connect to the DB
try:
    db=pgdb.connect(user=db_user,password=db_password,host=db_host,database=db_database)
    cursor=db.cursor()
except:
    sys.exit("Failed to connect to the DB")

def getItems(sql):
    "Returns an array of bugnumber for the given SQL queries"
    itemlist=set()
    for query in sql:
        cursor.execute(query)
        for item in cursor.fetchall():
            if item[0]:
                itemlist.add(item[0])
    return itemlist

def updateBug(bugnumber):
    "Get LP information about bugnumber, tag if a tag is specified and update the DB"

    bugnumber=int(bugnumber)
    try:
        bug=launchpad.bugs[bugnumber]
    except HTTPError:
        if verbose >= 1: print "bug #"+str(bugnumber)+": HTTP error"
        return
    except KeyError:
        if verbose >= 1: print "bug #"+str(bugnumber)+": Unknown bug"
        return
    except KeyboardInterrupt:
        sys.exit("KeyboardInterrupt caught.")
    except:
        if verbose >= 1: print "bug #"+str(bugnumber)+": Unhandled error when opening bug"
        return

    main=0
    queries=[]

    try:
        if bug.duplicate_of:
            mainbug=bug.duplicate_of.id
            bug=launchpad.bugs[mainbug]
        else:
            mainbug=bugnumber

        if isinstance(mainbug, int) and isinstance(bugnumber, int):
            bug_bugnumber=str(mainbug)
            bug_originalbug=str(bugnumber)

            if bug.title:
                bug_title=pgdb._quote(bug.title)
            else:
                bug_title="'None'"

            # 2009-10-01 - just use the first Ubuntu bug task found for lack of a better idea
            task_links = [task.self_link for task in bug.bug_tasks]
            task_links.sort(reverse=True)

            for task_link in task_links:
                task = launchpad.load(task_link)
                bug_product = task.bug_target_display_name
                bug_product=pgdb._quote(bug_product)
                bug_status=pgdb._quote(task.status)
                bug_importance=pgdb._quote(task.importance)
                if task.assignee:
                    try:
                        bug_assignee=pgdb._quote(task.assignee.display_name)
                    except:
                        bug_assignee=pgdb._quote(task.assignee.name)
                else:
                    bug_assignee=pgdb._quote("None")

            # FIXME: LP returns 500 for some bugs, workaround that for now.
            try:
                if len(bug.messages) > 0:
                    bug_comments=str(len(bug.messages))
                else:
                    bug_comments="0"
            except:
                bug_comments="0"

            # this used to be direct subscribers only but this doesn't exist in
            # python-launchpadlib
            if len(bug.subscriptions) > 0:
                bug_subscribers=str(len(bug.subscriptions))
            else:
                bug_subscribers="0"
            bug_duplicates=str(len(bug.duplicates))

            # If wanted, tag the bug
            if tag != 0:
                try:
                    cursor.execute("SELECT DISTINCT qawebsite_module_setting.value FROM qatracker_result LEFT JOIN qatracker_bug ON qatracker_bug.resultid=qatracker_result.id LEFT JOIN qatracker_build ON qatracker_build.id=qatracker_result.buildid LEFT JOIN qatracker_milestone ON qatracker_milestone.id=qatracker_build.milestoneid LEFT JOIN qawebsite_module_setting ON qawebsite_module_setting.siteid=qatracker_milestone.siteid WHERE qawebsite_module_setting.option='bug_tag' AND qawebsite_module_setting.option!='' AND qatracker_bug.bugnumber='"+str(bugnumber)+"'")
                    tags=cursor.fetchall()
                    for tagtext in tags:
                        if tagtext[0] != "" and tagtext[0] not in bug.tags:
                            taglist=bug.tags
                            taglist.append(unicode(str(tagtext[0])))
                            bug.tags=taglist
                            bug.lp_save()
                            if verbose >= 2: print "bug #"+str(bugnumber)+": Added \""+str(tagtext[0])+"\" tag"
                except KeyboardInterrupt:
                    sys.exit("KeyboardInterrupt caught.")
                except:
                    if verbose >= 1: print "bug #"+str(bugnumber)+": Unable to add \""+str(tagtext[0])+"\" tag"

            # Bug mentoring no longer exists, set to disable by default.
            bug_mentoring="0"
        else:
            if verbose >= 1: print "bug #"+str(bugnumber)+": Invalid bug object"
            return

    except KeyboardInterrupt:
        sys.exit("KeyboardInterrupt caught.")
    except HTTPError:
        if verbose >= 1: print "bug #"+str(bugnumber)+": HTTP error"
        return
    except:
        if verbose >= 1: print "bug #"+str(bugnumber)+": Error opening bug"
        return

    try:
        cursor.execute("SELECT originalbug FROM qawebsite_launchpad_bug WHERE originalbug='"+bug_originalbug+"'")
        if len(cursor.fetchall()) == 0:
            cursor.execute("INSERT INTO qawebsite_launchpad_bug (bugnumber, originalbug, title, product, status, importance, assignee, mentoring, commentscount, subscriberscount, duplicatescount) VALUES ("+bug_bugnumber+","+bug_originalbug+","+bug_title+","+bug_product+","+bug_status+","+bug_importance+","+bug_assignee+","+bug_mentoring+","+bug_comments+","+bug_subscribers+","+bug_duplicates+")")
        else:
            cursor.execute("UPDATE qawebsite_launchpad_bug SET lastchange=now() WHERE originalbug='"+bug_originalbug+"' AND status!="+bug_status)
            cursor.execute("UPDATE qawebsite_launchpad_bug SET bugnumber="+bug_bugnumber+", title="+bug_title+", product="+bug_product+", status="+bug_status+", importance="+bug_importance+", assignee="+bug_assignee+", mentoring="+bug_mentoring+", commentscount="+bug_comments+", subscriberscount="+bug_subscribers+", duplicatescount="+bug_duplicates+" WHERE originalbug='"+bug_originalbug+"'")
        db.commit()
        if verbose >= 3: print "bug #"+str(bugnumber)+": Successfully updated"
    except KeyboardInterrupt:
        sys.exit("KeyboardInterrupt caught.")
    except:
        if verbose >= 1: print "bug #"+str(bugnumber)+": DB faillure"
        return

# Arrays containing the SQL queries to execute
bug_new=[]
bug_current=[]
bug_archive=[]

# Check if we have the right qawebsite_launchpad_bug table
try:
    cursor.execute("SELECT bugnumber,originalbug,title,product,status,importance,assignee,mentoring,commentscount,subscriberscount,duplicatescount FROM qawebsite_launchpad_bug")
except:
    sys.exit("table qawebsite_launchpad_bug missing")

# Check what are the installed modules
try:
    cursor.execute("SELECT bugnumber FROM qatracker_bug")
    bug_new.append("SELECT DISTINCT bugnumber FROM qatracker_bug WHERE bugnumber NOT IN (SELECT originalbug FROM qawebsite_launchpad_bug)")
    bug_current.append("SELECT DISTINCT qatracker_bug.bugnumber FROM qatracker_result LEFT JOIN qatracker_build ON qatracker_result.buildid=qatracker_build.id LEFT JOIN qatracker_milestone ON qatracker_milestone.id=qatracker_build.milestoneid LEFT JOIN qatracker_bug ON qatracker_bug.resultid=qatracker_result.id WHERE qatracker_milestone.status!=2")
    bug_archive.append("SELECT DISTINCT qatracker_bug.bugnumber FROM qatracker_result LEFT JOIN qatracker_build ON qatracker_result.buildid=qatracker_build.id LEFT JOIN qatracker_milestone ON qatracker_milestone.id=qatracker_build.milestoneid LEFT JOIN qatracker_bug ON qatracker_bug.resultid=qatracker_result.id WHERE qatracker_milestone.status=2")
except KeyboardInterrupt:
    sys.exit("KeyboardInterrupt caught.")
except:
    db.rollback()
    if verbose >= 1: print "Module qatracker not found"

# Generate a set of bug to parse
if len(sys.argv) == 2:
    if sys.argv[1] == "update-new":
        if verbose >= 3: print "Selecting new items"
        bugs=getItems(bug_new)
    elif sys.argv[1] == "update-current":
        if verbose >= 3: print "Selecting currently shown items"
        bugs=getItems(bug_current)
    elif sys.argv[1] == "update-archive":
        if verbose >= 3: print "Selecting archived items"
        bugs=getItems(bug_archive)
    else:
        sys.exit("Unknown action \""+sys.argv[1]+"\"")
else:
    sys.exit("Syntax: "+sys.argv[0]+" <action>")

#Now parsing bugs
for bug in bugs:
    updateBug(bug)

#Close DB connection
cursor.close()
db.close()

if lock:
    os.remove("/tmp/lp-integration.lock")
